from collections import defaultdict
from enum import Enum

from pydantic import BaseModel


class _HashModel(BaseModel):
    def __hash__(self):
        return hash((type(self),) + tuple(self.__dict__.values()))

# Contests

statdict = defaultdict(lambda:2)
statdict[""] = 0
statdict["Pending"] = 1

class CStatus(Enum):
	approved = 0
	denied = 2
	pending = 1

# Results

class RStatus(Enum):
	OK = 0
	ANS = 1
	TLE = 2
	RTE = 3
	MEM = 4
	EXT = 5
	CME = 6
	INT = 7
	REJ = 8
	RUL = 9
	QUE = 10



from datetime import datetime, timedelta
from functools import lru_cache, wraps
from typing import List, Union

import markdownify as md
from lxml import etree, html

BASE_URL = "https://satori.tcs.uj.edu.pl"

def parse_news(response: str) -> List[str]:
	dom = html.fromstring(response)
	dom.make_links_absolute(BASE_URL)

	out = []
	for div in dom.xpath(r"//div[@id='content']/div"):
		for tr in div.xpath(r".//table[@class='message']"):
			content = etree.tostring(tr, encoding="unicode", with_tail=False)
			content = md.markdownify(content)
			lines = []
			for line in content.split("\n"):
				if line == "":
					continue
				if line == "> ":
					continue
				lines.append(line)

			content = "\n".join(lines)
			out.append(content)

	return out


def xpstr(tree, xpath:str) -> str:
	res = tree.xpath(xpath)
	res = "".join(res)
	return res


def timed_lru_cache(seconds:int=60, maxsize:Union[int,None]=128):
	def wrapper_cache(func):
		func = lru_cache(maxsize=maxsize)(func)
		setattr(func, "lifetime_", timedelta(seconds=seconds))
		setattr(func, "expiration_", datetime.utcnow() + getattr(func, "lifetime_"))

		@wraps(func)
		def wrapped_func(*args, **kwargs):
			if datetime.utcnow() >= getattr(func, "expiration_"):
				func.cache_clear()
				setattr(func, "expiration_", datetime.utcnow() + getattr(func, "lifetime_"))
			return func(*args, **kwargs)

		return wrapped_func

	return wrapper_cache




import re
from datetime import datetime
from functools import lru_cache
from itertools import chain
from typing import List
from urllib.error import HTTPError

import lxml
import lxml.html
import requests

from koishi.exceptions import AccessError
from koishi.models import CStatus, RStatus
from koishi.problem import Problem, ProblemSet
from koishi.result import Result
from koishi.utils import parse_news, xpstr, timed_lru_cache


class Contest:
	"""Wrapper class for satori contest.

	Args:
		ses: requests.Session
		id: int
		name: str
		desc: str
		status: CStatus

	Fields:
		Contest.id: int
		Contest.name: str
		Contest.desc: str
		Contest.status: CStatus

	Properties:
		news:
			A list of markdown formatted messages.
		problems:
			A list of problems sorted from oldest to newest.
		result:
			The result of the latest submit.
		results:
			A list of latest results.
	"""
	def __init__(self, ses: requests.Session, id: int, name: str, desc: str, status: CStatus):
		self._ses = ses
		self.id = id
		self.name = name
		self.desc = desc
		self.status = status

	@timed_lru_cache(7200, None)
	def __get_news(self) -> List[str]:
		resp = self._ses.get(f"https://satori.tcs.uj.edu.pl/contest/{self.id}/news")
		return parse_news(resp.text)

	@timed_lru_cache(7200, None)
	def __get_problemsets(self) -> List[ProblemSet]:
		if self.status.value > 1:
			raise AccessError(f"account not a member of \"{self.name}\" contest")

		resp = self._ses.get(f"https://satori.tcs.uj.edu.pl/contest/{self.id}/problems")
		dom = lxml.html.fromstring(resp.text)

		if resp.status_code == 404:
			raise HTTPError(resp.url,resp.status_code,"Contest not found",
				hdrs={},fp=None)

		out = []
		for series in zip(
			dom.xpath(r"//div[@id='content']/h4"),
			dom.xpath(r"//div[@id='content']/div")
		):
			title_, tasks_ = series
			title = xpstr(title_, ".//text()")
			title = re.sub(r"^\s+|\s+.$|\s*\[\w+\/\w+\]\s*$", "", title) # " {} [show/hide]  " -> "{}"

			tasks = []
			for tr in tasks_.xpath(r".//table[@class='results']/tr[position()!=1]"):
				code = xpstr(tr, r".//td[1]/text()")
				name = xpstr(tr, r".//td[2]/a/text()")
				note = xpstr(tr, r".//*[@class='mainsphinx']/p/em/text()")
				id  = xpstr(tr, r".//td[2]/a/@href")
				id  = re.sub(r"^\/\w+\/\d+\/\w+\/|\/$", "", id)

				tasks.append(
					Problem(self._ses, code=code, id=id, note=note, name=name, parent_id=self.id)
				)

			out.append(
				ProblemSet(title=title, problems=tasks)
			)

		return out


	@property
	def news(self) -> List[str]:
		"""A list of markdown formatted messages."""
		return self.__get_news()

	@property
	def problemsets(self) -> List[ProblemSet]:
		"""A list of ProblemSet objects sorted from oldest to newest"""
		return self.__get_problemsets()

	@property
	def problems(self) -> List[Problem]:
		"""A list of problems  sorted from oldest to newest"""
		return [*chain(*[x.problems for x in self.problemsets])]

	@property
	def result(self) -> Result:
		"""The latest submit result."""
		return self.results[0]

	@property
	def results(self) -> List[Result]:
		"""A list of latest results."""
		resp = self._ses.get(f"https://satori.tcs.uj.edu.pl/contest/{self.id}/results?results_limit=256")
		dom = lxml.html.fromstring(resp.text)

		out = []
		for tr in dom.xpath(r"//table[@class='results']/tr[position()!=1]"):
			id = xpstr(tr, r".//td[1]/a/text()")
			src = xpstr(tr, r".//td[2]/text()")

			date = xpstr(tr, r".//td[3]/text()")
			date = datetime.strptime(date, "%Y-%m-%d %H:%M:%S")

			status = xpstr(tr, r".//td[@class='status']/div[@class='submitstatus']/div/text()")
			status = RStatus[status]

			out.append(
				Result(self._ses, id, self.id, src, status, date)
			)

		return out

	def __hash__(self):
			return hash((type(self), self.id))

	def __repr__(self):
		return f"<Contest {self.id} {self.status.name} \"{self.name}\">"

	def __str__(self):
		return repr(self)

	def __eq__(self, other):
		return self.id == other.id

	def __ne__(self, other):
		return self.id != other.id

	def __getitem__(self, key):
		return self.problems[key]

	def __delitem__(self, key):
		del self.problems[key]

	def __setitem__(self, key, val):
		raise NotImplementedError(
			"Overriding problems is not supported"
		)




import re
from typing import Dict, List
from urllib.error import HTTPError

import lxml
import lxml.html
import requests

from koishi.contest import Contest
from koishi.models import CStatus, statdict
from koishi.utils import parse_news, xpstr, timed_lru_cache


class Koishi:
	""""""
	def __init__(self, login:str, password:str):
		self.__login = login
		self.__passwd = password
		self._ses = requests.Session() # Stores the session cookie

		resp = self._ses.post(
			"https://satori.tcs.uj.edu.pl/login",
			files = (
				("login", (None, self.__login)),
				("password", (None, self.__passwd))
			)
		)
		dom = lxml.html.fromstring(resp.text)

		if not resp.ok:
			raise HTTPError(resp.url,resp.status_code, "Internal satori error",
				hdrs={},fp=None)

		if dom.xpath(r"//*[@id='content']/form/table/tr[1]/td/text()='Login failed!'"):
			raise HTTPError(resp.url,resp.status_code, "Invalid login credentials",
				hdrs={},fp=None)

	@timed_lru_cache(7200, None)
	def _get_contests(self) -> List[Contest]:
		resp = self._ses.get("https://satori.tcs.uj.edu.pl/contest/select")
		dom = lxml.html.fromstring(resp.text)

		out = []
		for table in dom.xpath(r"//table[@class='results']"):
			for tr in table.xpath(r".//tr[position()!=1]"):
				id = xpstr(tr, r".//a[@class='stdlink']/@href")
				id = int(re.sub(r"^\/\w+\/|\/$", "", id))

				status = xpstr(tr, r".//td[3]//text()")
				status = CStatus(statdict[status])

				name = xpstr(tr, r".//a[@class='stdlink']/text()")
				desc = xpstr(tr, r".//td[@class='description']/text()")

				out.append(
					Contest(self._ses, id, name, desc, status)
				)

		if len(out) == 1 and out[0].id == 251519:
			raise HTTPError(resp.url,resp.status_code,
				"An error occurred while fetching user's contests",
				hdrs={},fp=None)

		# return [*sorted(out, key=lambda x,y: x.id > y.id)]
		return out

	@timed_lru_cache(7200, None)
	def __get_news(self) -> List[str]:
		resp = self._ses.get(f"https://satori.tcs.uj.edu.pl/news")
		return parse_news(resp.text)

	@property
	def contests(self) -> List[Contest]:
		"""Returns a list of user's contests"""
		return self._get_contests()

	@property
	def news(self) -> List[str]:
		"""Returns a list of Markdown formatted posts"""
		return self.__get_news()

	@property
	def profile(self) -> Dict[str, str]:
		"""Ret"""
		resp = self._ses.get("https://satori.tcs.uj.edu.pl/profile")
		dom = lxml.html.fromstring(resp.text)

		fname = xpstr(dom, r"//input[@id='id_firstname']/@value")
		lname = xpstr(dom, r"//input[@id='id_lastname']/@value")
		affi = xpstr(dom, r"//input[@id='id_affiliation']/@value")

		return {
			"first_name": fname,
			"last_name": lname,
			"affiliation": affi,
		}

	def relog(self) -> bool:
		"""Relogs user with credentials provided in `__init__`"""
		resp = self._ses.post(
			"https://satori.tcs.uj.edu.pl/login",
			files = (
				("login", (None, self.__login)),
				("password", (None, self.__passwd))
			)
		)
		return resp.ok

	def update(self, **fields) -> bool:
		"""Updates user's profile.

		Updates users satori profile with specified keyword arguments.
		When no value is provided, parameter's value will not be updated.

		Args:
			first_name:
				A string representing the new first name of the user.
			last_name:
				A string representing the new last name of the user.
			password:
				A string representing the new password of the user.
			affiliation:
				A string representing the new affiliation of the user.

		Raises:
			urllib.error.HTTPError: An error occurred while updating the profile.
		"""
		fname = self.profile["first_name"]
		lname = self.profile["last_name"]
		affi  = self.profile["affiliation"]

		fname = fields.get("first_name", fname)
		lname = fields.get("last_name", lname)
		affi = fields.get("affiliation", affi)

		newpass = fields.get("password", "")
		oldpass = self.__passwd if newpass else ""

		resp = self._ses.post(
			"https://satori.tcs.uj.edu.pl/profile",
			files=(
				("firstname", (None, fname)),
				("lastname", (None, lname)),
				("oldpass", (None, oldpass)),
				("newpass", (None, newpass)),
				("confirm", (None, newpass)),
				("affiliation", (None, affi)),
				("changepass", (None, "change"))
			)
		)

		if not resp.ok:
			raise HTTPError(resp.url,resp.status_code,
				"An error occurred while updating the profile",
				hdrs={},fp=None)

		return tuple(self.profile.values()) == (fname, lname, affi)

	def __hash__(self):
		return hash((type(self), self.__login, self.__passwd))

	def __repr__(self):
		return f"<{type(self)} #{hash(self)}>"

	def __str__(self):
		return repr(self)

	def __eq__(self, other):
		return hash(self) == hash(other)

	def __ne__(self, other):
		return hash(self) != hash(other)

	def __getitem__(self, key):
		return self.contests[key]

	def __delitem__(self, key):
		del self.contests[key]

	def __setitem__(self, key, val):
		raise NotImplementedError(
			"Overriding contests is not supported"
		)

	def __len__(self):
		return len(self.contests)

	def __call__(self):
		return self.relog()



class _ErrorFactory(Exception):
	def __init__(self, name:str, reason:str, *args):
		self.name = name
		self.reason = reason

	def __str__(self):
		return f"<Koishi {self.name} error, {self.reason}>"
	def __repr__(self):
		return str(self)


class LoginError(_ErrorFactory):
	def __init__(self, reason:str, *args):
		super().__init__("Login", reason, *args)

class AccessError(_ErrorFactory):
	def __init__(self, reason:str, *args):
		super().__init__("Permission", reason, *args)


import html as libhtml
import re
from datetime import datetime
from functools import lru_cache
from typing import List

import lxml
import lxml.html
import requests as rq

from koishi.models import RStatus, _HashModel
from koishi.utils import xpstr


class ResultTest(_HashModel):
	name:str
	status:RStatus
	time:str


class ResultDetails(_HashModel):
	"""Detailed version of class Result"""
	tests:List[ResultTest]
	source:str

	def __repr__(self):
		text = repr(self.source)[1:48]+"..." if len(self.source) > 50 else repr(self.source)
		test_val = [x.status.name for x in self.tests]
		test_val = " ".join(test_val)
		return f"<DetailedResult {test_val} {text}>"

	def __str__(self):
		return repr(self)


class Result:
	"""Wrapper class for satori problem result"""
	def __init__(self, ses:rq.Session, id:int, parent_id:int, src:str, status:RStatus, date:datetime):
		self._ses = ses
		self.id = id
		self.parent_id = parent_id
		self.src = src
		self.status = status
		self.date = date

	@lru_cache(None)
	def _get_details(self):
		resp = self._ses.get(f"https://satori.tcs.uj.edu.pl/contest/{self.parent_id}/results/{self.id}")
		dom = lxml.html.fromstring(resp.text)

		tests = []
		for tr in dom.xpath(r"//table[@class='docutils']/tbody[@valign='top']/tr"):
			name = xpstr(tr, r".//td[1]/text()")

			status = xpstr(tr, r".//td[2]/text()")
			status = RStatus[status]

			time = xpstr(tr, r".//td[3]/text()")
			time = re.sub(r"\D+$", "", time)

			tests.append(
				ResultTest(name=name, status=status, time=time)
			)

		source = xpstr(dom, r"//pre[@class='literal-block']/text()")
		source = libhtml.unescape(source)

		return ResultDetails(tests=tests, source=source)

	@property
	def details(self):
		"""Returns a ResultDetails object made from self"""
		return self._get_details()

	@property
	def ok(self):
		"""Returns True when result is positive"""
		return not self.status.value

	def __eq__(self, other):
		return self.id == other.id

	def __ne__(self, other):
		return not self.id == other.id

	def __hash__(self):
		return hash(self.id)

	def __repr__(self):
		return f"<Result {self.id} {self.src} {self.date} {self.status.name}>"

	def __str__(self):
		return repr(self)



import setuptools

setuptools.setup(
	name="koishi",
	version="1.1.4",
	description="Python wrapper for the unofficial scraped API of the satori testing system.",
	long_description=open("README.md", "r").read(),
	long_description_content_type="text/markdown",
	author="Cloud11665",
	author_email="Cloud11665@gmail.com",
	url="https://github.com/Cloud11665/koishi",
	packages=setuptools.find_packages(),
	install_requires=open("requirements.txt", "r").readlines(),
	license="MIT",
	keywords=["koishi", "tcs", "jagiellonian-university", "satori"],
	classifiers=[
		"License :: OSI Approved :: MIT License",
		"Programming Language :: Python :: 3",
		"Programming Language :: Python :: 3.8",
		"Programming Language :: Python :: 3.9",
		"Programming Language :: Python :: 3.10",
		"Operating System :: OS Independent",
	],
	python_requires=">=3.8" # will test with >=3.6
)


from typing import List, Any, Callable
from datetime import datetime
import time

import requests as rq
import lxml
import lxml.html

from koishi.utils import xpstr
from koishi.models import _HashModel
from koishi.models import RStatus
from koishi.result import Result


class ProblemSet(_HashModel):
	problems:List[Any] # fucking pydantic 'arbitrary_types_allowed' bullshit
	title:str

class Problem:
	"""Wrapper class for satori's problem"""
	def __init__(self, ses:rq.Session, id:int, parent_id:int, code:str, name:str, note:str):
		self._ses = ses
		self.id = id
		self.parent_id = parent_id
		self.code = code
		self.name = name
		self.note = note

	@property
	def pdf_url(self) -> str:
		return f"https://satori.tcs.uj.edu.pl/view/ProblemMapping/{self.parent_id}/statement_files/_pdf/{self.code}.pdf"

	def submit(self, src:str, lang:str, callback:Callable=None) -> int:
		resp = self._ses.post(
			f"https://satori.tcs.uj.edu.pl/contest/{self.parent_id}/submit",
			files = (
				("problem", (None, self.id)),
				("codefile", (f"sent_using_koishilib.{lang}", src))
			)
		)

		dom = lxml.html.fromstring(resp.text)
		tr = dom.xpath(r"//table[@class='results']/tr[2]")[0]
		id = int(xpstr(tr, r".//td[1]/a/text()"))

		date = xpstr(tr, r".//td[3]/text()")
		date = datetime.strptime(date, "%Y-%m-%d %H:%M:%S")

		status = xpstr(tr, r".//td[@class='status']/div[@class='submitstatus']/div/text()")
		status = RStatus[status]

		if callback:
			while status == RStatus.QUE:
				try:
					cresp = self._ses.get(f"https://satori.tcs.uj.edu.pl/contest/{self.parent_id}/results/{id}")

					cdom = lxml.html.fromstring(cresp.text)

					status = xpstr(cdom, r"//div[@id='content']/table/tr[2]/td[5]/text()")
					status = RStatus[status]

					time.sleep(5)
				except rq.exceptions.ConnectionError:
					... # :) your ratelimits are worthless..

			print(id, self.id, src, status, date)
			res = Result(self._ses, id, self.id, src, status, date)

			return callback(res)

		return id

	def __hash__(self):
		return hash((type(self),) + tuple(self.__dict__.values()))

	def __repr__(self):
		return f"<Problem {self.id} {self.code} {self.name}>"

	def __str__(self):
		return repr(self)

	def __eq__(self, other):
		return self.id == other.id

	def __ne__(self, other):
		return self.id != other.id

	def __call__(self, src:str, lang:str, callback:Callable=None):
		self.submit(src,lang,callback)